#include "defines.h"
#include "kozos.h"
#include "intr.h"
#include "interrupt.h"
#include "syscall.h"
#include "lib.h"

#define THREAD_NUM 6
#define THREAD_NAME_SIZE 15

/* åɡƥ */
typedef struct _kz_context {
  uint32 sp; /* åݥ */
} kz_context;

/* ȥ롦֥å(TCB) */
typedef struct _kz_thread {
  struct _kz_thread *next;
  char name[THREAD_NAME_SIZE + 1]; /* å̾ */
  char *stack;    /* å */

  struct { /* åɤΥȡå(thread_init())Ϥѥ᡼ */
    kz_func_t func; /* åɤΥᥤؿ */
    int argc;       /* åɤΥᥤؿϤ argc */
    char **argv;    /* åɤΥᥤؿϤ argv */
  } init;

  struct { /* ƥࡦѥХåե */
    kz_syscall_type_t type;
    kz_syscall_param_t *param;
  } syscall;

  kz_context context; /* ƥȾ */
} kz_thread;

/* åɤΥǥ塼 */
static struct {
  kz_thread *head;
  kz_thread *tail;
} readyque;

static kz_thread *current; /* ȡå */
static kz_thread threads[THREAD_NUM]; /* ȥ롦֥å */
static kz_handler_t handlers[SOFTVEC_TYPE_NUM]; /* ߥϥɥ */

void dispatch(kz_context *context);

/* ȡåɤǥ塼ȴФ */
static int getcurrent(void)
{
  if (current == NULL) {
    return -1;
  }

  /* ȡåɤɬƬˤϤʤΤǡƬȴФ */
  readyque.head = current->next;
  if (readyque.head == NULL) {
    readyque.tail = NULL;
  }
  current->next = NULL;

  return 0;
}

/* ȡåɤǥ塼˷Ҥ */
static int putcurrent(void)
{
  if (current == NULL) {
    return -1;
  }

  /* ǥ塼³ */
  if (readyque.tail) {
    readyque.tail->next = current;
  } else {
    readyque.head = current;
  }
  readyque.tail = current;

  return 0;
}

static void thread_end(void)
{
  kz_exit();
}

/* åɤΥȡå */
static void thread_init(kz_thread *thp)
{
  /* åɤΥᥤؿƤӽФ */
  thp->init.func(thp->init.argc, thp->init.argv);
  thread_end();
}

/* ƥࡦν(kz_run():åɤεư) */
static kz_thread_id_t thread_run(kz_func_t func, char *name,
				 int stacksize, int argc, char *argv[])
{
  int i;
  kz_thread *thp;
  uint32 *sp;
  extern char userstack; /* 󥫡ץȤ륹åΰ */
  static char *thread_stack = &userstack;

  /* Ƥ륿ȥ롦֥å򸡺 */
  for (i = 0; i < THREAD_NUM; i++) {
    thp = &threads[i];
    if (!thp->init.func) /* Ĥä */
      break;
  }
  if (i == THREAD_NUM) /* Ĥʤä */
    return -1;

  memset(thp, 0, sizeof(*thp));

  /* ȥ롦֥å(TCB) */
  strcpy(thp->name, name);
  thp->next     = NULL;

  thp->init.func = func;
  thp->init.argc = argc;
  thp->init.argv = argv;

  /* åΰ */
  memset(thread_stack, 0, stacksize);
  thread_stack += stacksize;

  thp->stack = thread_stack; /* å */

  /* åν */
  sp = (uint32 *)thp->stack;
  *(--sp) = (uint32)thread_end;

  /*
   * ץࡦ󥿤ꤹ롥
   */
  *(--sp) = (uint32)thread_init;

  *(--sp) = 0; /* ER6 */
  *(--sp) = 0; /* ER5 */
  *(--sp) = 0; /* ER4 */
  *(--sp) = 0; /* ER3 */
  *(--sp) = 0; /* ER2 */
  *(--sp) = 0; /* ER1 */

  /* åɤΥȡå(thread_init())Ϥ */
  *(--sp) = (uint32)thp;  /* ER0 */

  /* åɤΥƥȤ */
  thp->context.sp = (uint32)sp;

  /* ƥࡦƤӽФåɤǥ塼᤹ */
  putcurrent();

  /* åɤ򡤥ǥ塼³ */
  current = thp;
  putcurrent();

  return (kz_thread_id_t)current;
}

/* ƥࡦν(kz_exit():åɤνλ) */
static int thread_exit(void)
{
  /*
   * ʤ饹åƺѤǤ褦ˤ٤ά
   * Τᡤåɤˤõ褦ʤȤϸǤǤʤ
   */
  puts(current->name);
  puts(" EXIT.\n");
  memset(current, 0, sizeof(*current));
  return 0;
}

/* ߥϥɥϿ */
static int setintr(softvec_type_t type, kz_handler_t handler)
{
  static void thread_intr(softvec_type_t type, unsigned long sp);

  /*
   * ߤդ뤿ˡեȥߥ٥
   * OSγ߽ȤʤؿϿ롥
   */
  softvec_setintr(type, thread_intr);

  handlers[type] = handler; /* OS¦ƤӽФߥϥɥϿ */

  return 0;
}

static void call_functions(kz_syscall_type_t type, kz_syscall_param_t *p)
{
  /* ƥࡦμ¹current񤭴Τ */
  switch (type) {
  case KZ_SYSCALL_TYPE_RUN: /* kz_run() */
    p->un.run.ret = thread_run(p->un.run.func, p->un.run.name,
			       p->un.run.stacksize,
			       p->un.run.argc, p->un.run.argv);
    break;
  case KZ_SYSCALL_TYPE_EXIT: /* kz_exit() */
    /* TCBõΤǡͤ񤭹ǤϤʤ */
    thread_exit();
    break;
  default:
    break;
  }
}

/* ƥࡦν */
static void syscall_proc(kz_syscall_type_t type, kz_syscall_param_t *p)
{
  /*
   * ƥࡦƤӽФåɤǥ塼
   * ֤ǽؿƤӽФΤ᥷ƥࡦ
   * ƤӽФåɤ򤽤Τޤư³ˤϡ
   * ؿ putcurrent() Ԥɬפ롥
   */
  getcurrent();
  call_functions(type, p);
}

/* åɤΥ塼 */
static void schedule(void)
{
  if (!readyque.head) /* Ĥʤä */
    kz_sysdown();

  current = readyque.head; /* ȡåɤꤹ */
}

static void syscall_intr(void)
{
  syscall_proc(current->syscall.type, current->syscall.param);
}

static void softerr_intr(void)
{
  puts(current->name);
  puts(" DOWN.\n");
  getcurrent(); /* ǥ塼鳰 */
  thread_exit(); /* åɽλ */
}

/* ߽ؿ */
static void thread_intr(softvec_type_t type, unsigned long sp)
{
  /* ȡåɤΥƥȤ¸ */
  current->context.sp = sp;

  /*
   * ߤȤν¹Ԥ롥
   * SOFTVEC_TYPE_SYSCALL, SOFTVEC_TYPE_SOFTERR ξ
   * syscall_intr(), softerr_intr() ϥɥϿƤΤǡ
   * 餬¹Ԥ롥
   */
  if (handlers[type])
    handlers[type]();

  schedule(); /* åɤΥ塼 */

  /*
   * åɤΥǥѥå
   * (dispatch()ؿΤstartup.sˤꡤ֥ǵҤƤ)
   */
  dispatch(&current->context);
  /* ˤ֤äƤʤ */
}

void kz_start(kz_func_t func, char *name, int stacksize,
	      int argc, char *argv[])
{
  /*
   * ʹߤǸƤӽФåɴϢΥ饤֥ؿ current 
   * Ƥ礬Τǡcurrent  NULL ˽Ƥ
   */
  current = NULL;

  readyque.head = readyque.tail = NULL;
  memset(threads,  0, sizeof(threads));
  memset(handlers, 0, sizeof(handlers));

  /* ߥϥɥϿ */
  setintr(SOFTVEC_TYPE_SYSCALL, syscall_intr); /* ƥࡦ */
  setintr(SOFTVEC_TYPE_SOFTERR, softerr_intr); /* װȯ */

  /* ƥࡦȯԲĤʤΤľܴؿƤӽФƥåɺ */
  current = (kz_thread *)thread_run(func, name, stacksize, argc, argv);

  /* ǽΥåɤư */
  dispatch(&current->context);

  /* ˤ֤äƤʤ */
}

void kz_sysdown(void)
{
  puts("system error!\n");
  while (1)
    ;
}

/* ƥࡦƤӽФѥ饤֥ؿ */
void kz_syscall(kz_syscall_type_t type, kz_syscall_param_t *param)
{
  current->syscall.type  = type;
  current->syscall.param = param;
  asm volatile ("trapa #0"); /* ȥå׳ȯ */
}
